@@ -1,5 +1,4 @@
-#
-# Copyright (C) 2008, 2009 Canonical Ltd.
+# Copyright (C) 2006-2011 Canonical Ltd.
 #                     (Authored by Martin Albisetti <argentina@gmail.com>)
 # Copyright (C) 2006  Robey Pointer <robey@lag.net>
 # Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
@@ -18,7 +17,7 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
 #
 
 #
@@ -35,22 +34,24 @@ import re
 import textwrap
 import threading
 
-import bzrlib.branch
-import bzrlib.delta
-import bzrlib.errors
-import bzrlib.foreign
-import bzrlib.revision
+from breezy import tag
+import breezy.branch
+import breezy.delta
+import breezy.errors
+import breezy.foreign
+import breezy.osutils
+import breezy.revision
 
-from loggerhead import search
-from loggerhead import util
-from loggerhead.wholehistory import compute_whole_history_data
+from . import search
+from . import util
+from .wholehistory import compute_whole_history_data
 
 
 def is_branch(folder):
     try:
-        bzrlib.branch.Branch.open(folder)
+        breezy.branch.Branch.open(folder)
         return True
-    except:
+    except breezy.errors.NotBranchError:
         return False
 
 
@@ -105,57 +106,65 @@ class _RevListToTimestamps(object):
     def __len__(self):
         return len(self.revid_list)
 
+
 class FileChangeReporter(object):
 
-    def __init__(self, old_inv, new_inv):
+    def __init__(self, old_tree, new_tree):
         self.added = []
         self.modified = []
         self.renamed = []
         self.removed = []
         self.text_changes = []
-        self.old_inv = old_inv
-        self.new_inv = new_inv
+        self.old_tree = old_tree
+        self.new_tree = new_tree
 
-    def revid(self, inv, file_id):
+    def revid(self, tree, path):
+        if path is None:
+            return breezy.revision.NULL_REVISION
         try:
-            return inv[file_id].revision
-        except bzrlib.errors.NoSuchId:
-            return 'null:'
+            return tree.get_file_revision(path)
+        except breezy.errors.NoSuchFile:
+            return breezy.revision.NULL_REVISION
 
-    def report(self, file_id, paths, versioned, renamed, modified,
+    def report(self, paths, versioned, renamed, copied, modified,
                exe_change, kind):
+        return self._report(
+                paths, versioned, renamed, copied,
+                modified, exe_change, kind)
+
+    def _report(self, paths, versioned, renamed, copied, modified,
+                exe_change, kind):
         if modified not in ('unchanged', 'kind changed'):
             if versioned == 'removed':
                 filename = rich_filename(paths[0], kind[0])
             else:
                 filename = rich_filename(paths[1], kind[1])
             self.text_changes.append(util.Container(
-                filename=filename, file_id=file_id,
-                old_revision=self.revid(self.old_inv, file_id),
-                new_revision=self.revid(self.new_inv, file_id)))
+                filename=filename,
+                old_revision=self.revid(self.old_tree, paths[0]),
+                new_revision=self.revid(self.new_tree, paths[1])))
         if versioned == 'added':
             self.added.append(util.Container(
-                filename=rich_filename(paths[1], kind),
-                file_id=file_id, kind=kind[1]))
+                filename=rich_filename(paths[1], kind), kind=kind[1]))
         elif versioned == 'removed':
             self.removed.append(util.Container(
-                filename=rich_filename(paths[0], kind),
-                file_id=file_id, kind=kind[0]))
+                filename=rich_filename(paths[0], kind), kind=kind[0]))
         elif renamed:
             self.renamed.append(util.Container(
                 old_filename=rich_filename(paths[0], kind[0]),
                 new_filename=rich_filename(paths[1], kind[1]),
-                file_id=file_id,
-                text_modified=modified == 'modified'))
+                text_modified=modified == 'modified', exe_change=exe_change))
         else:
             self.modified.append(util.Container(
                 filename=rich_filename(paths[1], kind),
-                file_id=file_id))
+                text_modified=modified == 'modified', exe_change=exe_change))
+
 
 # The lru_cache is not thread-safe, so we need a lock around it for
 # all threads.
 rev_info_memory_cache_lock = threading.RLock()
 
+
 class RevInfoMemoryCache(object):
     """A store that validates values against the revids they were stored with.
 
@@ -200,7 +209,7 @@ class RevInfoMemoryCache(object):
         finally:
             rev_info_memory_cache_lock.release()
 
-# Used to store locks that prevent multiple threads from building a 
+# Used to store locks that prevent multiple threads from building a
 # revision graph for the same branch at the same time, because that can
 # cause severe performance issues that are so bad that the system seems
 # to hang.
@@ -215,8 +224,6 @@ class History(object):
     around it, serve the request, throw the History object away, unlock the
     branch and throw it away.
 
-    :ivar _file_change_cache: An object that caches information about the
-        files that changed between two revisions.
     :ivar _rev_info: A list of information about revisions.  This is by far
         the most cryptic data structure in loggerhead.  At the top level, it
         is a list of 3-tuples [(merge-info, where-merged, parents)].
@@ -284,15 +291,10 @@ class History(object):
                 self._rev_indices[revid] = seq
                 self._revno_revid[revno_str] = revid
 
-    def __init__(self, branch, whole_history_data_cache, file_cache=None,
+    def __init__(self, branch, whole_history_data_cache,
                  revinfo_disk_cache=None, cache_key=None):
         assert branch.is_locked(), (
             "Can only construct a History object with a read-locked branch.")
-        if file_cache is not None:
-            self._file_change_cache = file_cache
-            file_cache.history = self
-        else:
-            self._file_change_cache = None
         self._branch = branch
         self._branch_tags = None
         self._inventory_cache = {}
@@ -308,7 +310,7 @@ class History(object):
 
     @property
     def has_revisions(self):
-        return not bzrlib.revision.is_null(self.last_revid)
+        return not breezy.revision.is_null(self.last_revid)
 
     def get_config(self):
         return self._branch.get_config()
@@ -327,7 +329,16 @@ class History(object):
         revid in revid_list.
         """
         if revid_list is None:
-            revid_list = [r[0][1] for r in self._rev_info]
+            # Just yield the mainline, starting at start_revid
+            revid = start_revid
+            is_null = breezy.revision.is_null
+            while not is_null(revid):
+                yield revid
+                parents = self._rev_info[self._rev_indices[revid]][2]
+                if not parents:
+                    return
+                revid = parents[0]
+            return
         revid_set = set(revid_list)
         revid = start_revid
 
@@ -340,10 +351,14 @@ class History(object):
                 r.add(self._rev_info[i][0][1])
                 i += 1
             return r
-        while True:
-            if bzrlib.revision.is_null(revid):
+        while revid_set:
+            if breezy.revision.is_null(revid):
                 return
-            if introduced_revisions(revid) & revid_set:
+            rev_introduced = introduced_revisions(revid)
+            matching = rev_introduced.intersection(revid_set)
+            if matching:
+                # We don't need to look for these anymore.
+                revid_set.difference_update(matching)
                 yield revid
             parents = self._rev_info[self._rev_indices[revid]][2]
             if len(parents) == 0:
@@ -353,14 +368,15 @@ class History(object):
     def get_short_revision_history_by_fileid(self, file_id):
         # FIXME: would be awesome if we could get, for a folder, the list of
         # revisions where items within that folder changed.i
+        # TODO(jelmer): Avoid versionedfile-specific texts
         possible_keys = [(file_id, revid) for revid in self._rev_indices]
         get_parent_map = self._branch.repository.texts.get_parent_map
         # We chunk the requests as this works better with GraphIndex.
-        # See _filter_revisions_touching_file_id in bzrlib/log.py
+        # See _filter_revisions_touching_file_id in breezy/log.py
         # for more information.
         revids = []
         chunk_size = 1000
-        for start in xrange(0, len(possible_keys), chunk_size):
+        for start in range(0, len(possible_keys), chunk_size):
             next_keys = possible_keys[start:start + chunk_size]
             revids += [k[1] for k in get_parent_map(next_keys)]
         del possible_keys, next_keys
@@ -401,8 +417,6 @@ iso style "yyyy-mm-dd")
         # ignore the passed-in revid_list
         revid = self.fix_revid(query)
         if revid is not None:
-            if isinstance(revid, unicode):
-                revid = revid.encode('utf-8')
             changes = self.get_changes([revid])
             if (changes is not None) and (len(changes) > 0):
                 return [revid]
@@ -444,16 +458,46 @@ iso style "yyyy-mm-dd")
         # if a "revid" is actually a dotted revno, convert it to a revid
         if revid is None:
             return revid
+        if not isinstance(revid, str):
+            raise TypeError(revid)
         if revid == 'head:':
             return self.last_revid
         try:
             if self.revno_re.match(revid):
                 revid = self._revno_revid[revid]
         except KeyError:
-            raise bzrlib.errors.NoSuchRevision(self._branch_nick, revid)
+            raise breezy.errors.NoSuchRevision(self._branch_nick, revid)
+        if not isinstance(revid, bytes):
+            revid = revid.encode('utf-8')
         return revid
 
-    def get_file_view(self, revid, file_id):
+    @staticmethod
+    def _iterate_sufficiently(iterable, stop_at, extra_rev_count):
+        """Return a list of iterable.
+
+        If extra_rev_count is None, fully consume iterable.
+        Otherwise, stop at 'stop_at' + extra_rev_count.
+
+        Example:
+          iterate until you find stop_at, then iterate 10 more times.
+        """
+        if extra_rev_count is None:
+            return list(iterable)
+        result = []
+        found = False
+        for n in iterable:
+            result.append(n)
+            if n == stop_at:
+                found = True
+                break
+        if found:
+            for count, n in enumerate(iterable):
+                if count >= extra_rev_count:
+                    break
+                result.append(n)
+        return result
+
+    def _get_file_view(self, revid, file_id):
         """
         Given a revid and optional path, return a (revlist, revid) for
         navigation through the current scope: from the revid (or the latest
@@ -464,25 +508,29 @@ iso style "yyyy-mm-dd")
         if revid is None:
             revid = self.last_revid
         if file_id is not None:
-            # since revid is 'start_revid', possibly should start the path
-            # tracing from revid... FIXME
-            revlist = list(self.get_short_revision_history_by_fileid(file_id))
-            revlist = list(self.get_revids_from(revlist, revid))
+            revlist = list(
+                self.get_short_revision_history_by_fileid(file_id))
+            revlist = self.get_revids_from(revlist, revid)
         else:
-            revlist = list(self.get_revids_from(None, revid))
+            revlist = self.get_revids_from(None, revid)
         return revlist
 
-    def get_view(self, revid, start_revid, file_id, query=None):
+    def get_view(self, revid, start_revid, path, query=None,
+                 extra_rev_count=None):
         """
-        use the URL parameters (revid, start_revid, file_id, and query) to
-        determine the revision list we're viewing (start_revid, file_id, query)
+        use the URL parameters (revid, start_revid, path, and query) to
+        determine the revision list we're viewing (start_revid, path, query)
         and where we are in it (revid).
 
             - if a query is given, we're viewing query results.
-            - if a file_id is given, we're viewing revisions for a specific
+            - if a path is given, we're viewing revisions for a specific
               file.
             - if a start_revid is given, we're viewing the branch from a
               specific revision up the tree.
+            - if extra_rev_count is given, find the view from start_revid =>
+              revid, and continue an additional 'extra_rev_count'. If not
+              given, then revid_list will contain the full history of
+              start_revid
 
         these may be combined to view revisions for a specific file, from
         a specific revision, with a specific search query.
@@ -493,26 +541,37 @@ iso style "yyyy-mm-dd")
             - start_revid: starting revision of this view
             - revid_list: list of revision ids for this view
 
-        file_id and query are never changed so aren't returned, but they may
+        path and query are never changed so aren't returned, but they may
         contain vital context for future url navigation.
         """
         if start_revid is None:
             start_revid = self.last_revid
 
         if query is None:
-            revid_list = self.get_file_view(start_revid, file_id)
+            repo = self._branch.repository
+            if path is not None:
+                file_id = repo.revision_tree(start_revid).path2id(path)
+            else:
+                file_id = None
+            revid_list = self._get_file_view(start_revid, file_id)
+            revid_list = self._iterate_sufficiently(revid_list, revid,
+                                                    extra_rev_count)
             if revid is None:
                 revid = start_revid
             if revid not in revid_list:
                 # if the given revid is not in the revlist, use a revlist that
                 # starts at the given revid.
-                revid_list = self.get_file_view(revid, file_id)
+                revid_list = self._get_file_view(revid, file_id)
+                revid_list = self._iterate_sufficiently(revid_list, revid,
+                                                        extra_rev_count)
                 start_revid = revid
             return revid, start_revid, revid_list
+        else:
+            file_id = None
 
         # potentially limit the search
         if file_id is not None:
-            revid_list = self.get_file_view(start_revid, file_id)
+            revid_list = self._get_file_view(start_revid, file_id)
         else:
             revid_list = None
         revid_list = search.search_revisions(self._branch, query)
@@ -526,24 +585,16 @@ iso style "yyyy-mm-dd")
             # search index.
             return None, None, []
 
-    def get_inventory(self, revid):
-        if revid not in self._inventory_cache:
-            self._inventory_cache[revid] = (
-                self._branch.repository.get_inventory(revid))
-        return self._inventory_cache[revid]
+    def revision_tree(self, revid):
+        return self._branch.repository.revision_tree(revid)
 
-    def get_path(self, revid, file_id):
-        if (file_id is None) or (file_id == ''):
-            return ''
-        path = self.get_inventory(revid).id2path(file_id)
+    def file_exists(self, revid, path):
         if (len(path) > 0) and not path.startswith('/'):
             path = '/' + path
-        return path
-
-    def get_file_id(self, revid, path):
-        if (len(path) > 0) and not path.startswith('/'):
-            path = '/' + path
-        return self.get_inventory(revid).path2id(path)
+        try:
+            return self.revision_tree(revid).has_filename(path)
+        except breezy.errors.NoSuchRevision:
+            return False
 
     def get_merge_point_list(self, revid):
         """
@@ -553,27 +604,17 @@ iso style "yyyy-mm-dd")
             return []
 
         merge_point = []
-        while True:
+        nexts = [revid]
+        while nexts:
+            revid = nexts.pop()
             children = self._rev_info[self._rev_indices[revid]][1]
-            nexts = []
             for child in children:
                 child_parents = self._rev_info[self._rev_indices[child]][2]
                 if child_parents[0] == revid:
                     nexts.append(child)
                 else:
                     merge_point.append(child)
-
-            if len(nexts) == 0:
-                # only merge
-                return merge_point
-
-            while len(nexts) > 1:
-                # branch
-                next = nexts.pop()
-                merge_point_next = self.get_merge_point_list(next)
-                merge_point.extend(merge_point_next)
-
-            revid = nexts[0]
+        return merge_point
 
     def simplify_merge_point_list(self, revids):
         """if a revision is already merged, don't show further merge points"""
@@ -590,7 +631,7 @@ iso style "yyyy-mm-dd")
             else:
                 d[revnos] = (revnolast, revid)
 
-        return [revid for (_, revid) in d.itervalues()]
+        return [revid for (_, revid) in d.values()]
 
     def add_branch_nicks(self, change):
         """
@@ -603,7 +644,7 @@ iso style "yyyy-mm-dd")
         for p in change.merge_points:
             fetch_set.add(p.revid)
         p_changes = self.get_changes(list(fetch_set))
-        p_change_dict = dict([(c.revid, c) for c in p_changes])
+        p_change_dict = {c.revid: c for c in p_changes}
         for p in change.parents:
             if p.revid in p_change_dict:
                 p.branch_nick = p_change_dict[p.revid].branch_nick
@@ -620,6 +661,9 @@ iso style "yyyy-mm-dd")
 
         Revisions not present and NULL_REVISION will be ignored.
         """
+        for revid in revid_list:
+            if not isinstance(revid, bytes):
+                raise TypeError(revid_list)
         changes = self.get_changes_uncached(revid_list)
         if len(changes) == 0:
             return changes
@@ -646,8 +690,8 @@ iso style "yyyy-mm-dd")
 
     def get_changes_uncached(self, revid_list):
         # FIXME: deprecated method in getting a null revision
-        revid_list = filter(lambda revid: not bzrlib.revision.is_null(revid),
-                            revid_list)
+        revid_list = list(filter(lambda revid: not breezy.revision.is_null(revid),
+                            revid_list))
         parent_map = self._branch.repository.get_graph().get_parent_map(
                          revid_list)
         # We need to return the answer in the same order as the input,
@@ -660,7 +704,7 @@ iso style "yyyy-mm-dd")
 
     def _change_from_revision(self, revision):
         """
-        Given a bzrlib Revision, return a processed "change" for use in
+        Given a breezy Revision, return a processed "change" for use in
         templates.
         """
         message, short_message = clean_message(revision.message)
@@ -670,12 +714,22 @@ iso style "yyyy-mm-dd")
 
         revtags = None
         if revision.revision_id in self._branch_tags:
-          revtags = ', '.join(self._branch_tags[revision.revision_id])
+          # tag.sort_* functions expect (tag, data) pairs, so we generate them,
+          # and then strip them
+          tags = [(t, None) for t in self._branch_tags[revision.revision_id]]
+          sort_func = getattr(tag, 'sort_natural', None)
+          if sort_func is None:
+              tags.sort()
+          else:
+              sort_func(self._branch, tags)
+          revtags = u', '.join([t[0] for t in tags])
 
         entry = {
             'revid': revision.revision_id,
             'date': datetime.datetime.fromtimestamp(revision.timestamp),
             'utc_date': datetime.datetime.utcfromtimestamp(revision.timestamp),
+            'timestamp': revision.timestamp,
+            'committer': revision.committer,
             'authors': revision.get_apparent_authors(),
             'branch_nick': revision.properties.get('branch-nick', None),
             'short_comment': short_message,
@@ -685,14 +739,15 @@ iso style "yyyy-mm-dd")
             'bugs': [bug.split()[0] for bug in revision.properties.get('bugs', '').splitlines()],
             'tags': revtags,
         }
-        if isinstance(revision, bzrlib.foreign.ForeignRevision):
-            foreign_revid, mapping = (rev.foreign_revid, rev.mapping)
-        elif ":" in revision.revision_id:
+        if isinstance(revision, breezy.foreign.ForeignRevision):
+            foreign_revid, mapping = (
+                revision.foreign_revid, revision.mapping)
+        elif b":" in revision.revision_id:
             try:
                 foreign_revid, mapping = \
-                    bzrlib.foreign.foreign_vcs_registry.parse_revision_id(
+                    breezy.foreign.foreign_vcs_registry.parse_revision_id(
                         revision.revision_id)
-            except bzrlib.errors.InvalidRevisionId:
+            except breezy.errors.InvalidRevisionId:
                 foreign_revid = None
                 mapping = None
         else:
@@ -702,61 +757,58 @@ iso style "yyyy-mm-dd")
             entry["foreign_revid"] = mapping.vcs.show_foreign_revid(foreign_revid)
         return util.Container(entry)
 
-    def get_file_changes_uncached(self, entry):
+    def get_file_changes(self, entry):
         if entry.parents:
             old_revid = entry.parents[0].revid
         else:
-            old_revid = bzrlib.revision.NULL_REVISION
+            old_revid = breezy.revision.NULL_REVISION
         return self.file_changes_for_revision_ids(old_revid, entry.revid)
 
-    def get_file_changes(self, entry):
-        if self._file_change_cache is None:
-            return self.get_file_changes_uncached(entry)
-        else:
-            return self._file_change_cache.get_file_changes(entry)
-
     def add_changes(self, entry):
         changes = self.get_file_changes(entry)
         entry.changes = changes
 
-    def get_file(self, file_id, revid):
+    def get_file(self, path, revid):
         """Returns (path, filename, file contents)"""
-        inv = self.get_inventory(revid)
-        inv_entry = inv[file_id]
-        rev_tree = self._branch.repository.revision_tree(inv_entry.revision)
-        path = inv.id2path(file_id)
-        if not path.startswith('/'):
+        if not isinstance(path, str):
+            raise TypeError(path)
+        if not isinstance(revid, bytes):
+            raise TypeError(revid)
+        rev_tree = self._branch.repository.revision_tree(revid)
+        display_path = path
+        if not display_path.startswith('/'):
             path = '/' + path
-        return path, inv_entry.name, rev_tree.get_file_text(file_id)
+        return (display_path, breezy.osutils.basename(path),
+                rev_tree.get_file_text(path))
 
     def file_changes_for_revision_ids(self, old_revid, new_revid):
         """
         Return a nested data structure containing the changes in a delta::
 
-            added: list((filename, file_id)),
-            renamed: list((old_filename, new_filename, file_id)),
-            deleted: list((filename, file_id)),
-            modified: list(
-                filename: str,
-                file_id: str,
-            ),
-            text_changes: list((filename, file_id)),
+            added: list((filename)),
+            renamed: list((old_filename, new_filename)),
+            deleted: list((filename)),
+            modified: list((filename)),
+            text_changes: list((filename)),
         """
         repo = self._branch.repository
-        if (bzrlib.revision.is_null(old_revid) or
-            bzrlib.revision.is_null(new_revid)):
+        if (breezy.revision.is_null(old_revid) or
+                breezy.revision.is_null(new_revid) or
+                old_revid == new_revid):
             old_tree, new_tree = map(
                 repo.revision_tree, [old_revid, new_revid])
         else:
             old_tree, new_tree = repo.revision_trees([old_revid, new_revid])
 
-        reporter = FileChangeReporter(old_tree.inventory, new_tree.inventory)
+        reporter = FileChangeReporter(old_tree, new_tree)
 
-        bzrlib.delta.report_changes(new_tree.iter_changes(old_tree), reporter)
+        breezy.delta.report_changes(new_tree.iter_changes(old_tree), reporter)
 
         return util.Container(
             added=sorted(reporter.added, key=lambda x: x.filename),
             renamed=sorted(reporter.renamed, key=lambda x: x.new_filename),
             removed=sorted(reporter.removed, key=lambda x: x.filename),
             modified=sorted(reporter.modified, key=lambda x: x.filename),
-            text_changes=sorted(reporter.text_changes, key=lambda x: x.filename))
+            text_changes=sorted(reporter.text_changes,
+                                key=lambda x: x.filename))
+
